import json
import os

import boto3
import pytest
import moto
from moto.ec2 import utils as ec2_utils


BASE_DIR = os.path.dirname(os.path.abspath(__file__))


@pytest.yield_fixture(scope="function")
def cluster():
    boto3.setup_default_session(region_name='eu-west-1')

    moto.mock_ecs().start()
    moto.mock_ec2().start()

    ec2 = boto3.resource('ec2', region_name='eu-west-1')
    ecs = boto3.client('ecs', region_name='eu-west-1')

    test_instance = ec2.create_instances(
        ImageId="ami-1234abcd",
        MinCount=1,
        MaxCount=1,
    )[0]

    instance_id_document = json.dumps(
        ec2_utils.generate_instance_identity_document(test_instance))

    cluster = ecs.create_cluster(clusterName='default')
    ecs.register_container_instance(
        cluster='default',
        instanceIdentityDocument=instance_id_document)

    yield cluster

    moto.mock_ecs().stop()
    moto.mock_ec2().stop()


@pytest.fixture
def definition():
    path = os.path.join(BASE_DIR, 'files/default_taskdef.json')

    with open(path, 'r') as json_file:
        return json.load(json_file)


@pytest.fixture
def default_config():
    path = os.path.join(BASE_DIR, 'files/default_config.yml')

    with open(path, 'r') as fh:
        yield fh



from ecs_deplojo.main import transform_definition
from ecs_deplojo import utils


def test_definition(definition):
    definition['containerDefinitions'][0]['environment'] = {
        'DEBUG': True,
        'AWS_REGION': 'eu-west-1',
    }
    result = transform_definition(definition)

    assert result['containerDefinitions'][0]['environment'] == [
        {
            'name': 'AWS_REGION',
            'value': 'eu-west-1',
        },
        {
            'name': 'DEBUG',
            'value': 'True',
        },
    ]


# def test_describe_services(connection):
#     task_definition = {
#         'family': 'my-task-def',
#         'volumes': [],
#         'containerDefinitions': [
#             {
#                 'name': 'default',
#                 'image': 'my-docker-image:1.0',
#                 'essential': True,
#                 'command': ['hello', 'world'],
#                 'memory': 256,
#                 'cpu': 0,
#                 'portMappings': [],
#                 'environment': {}
#             }
#         ]
#     }

#     connection.ecs.create_cluster(clusterName='default')
#     connection.ecs.create_service(
#         cluster='default',
#         serviceName='my-service',
#         desiredCount=1,
#         taskDefinition=task_definition)

#     result = utils.describe_services(
#         connection.ecs,
#         'default',
#         services=['my-service'])

#     assert result == ['1']



import os

from ecs_deplojo import task_definitions

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def test_generate_task_definition(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "default",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()

    filename = tmpdir.join('task_definition.json')
    filename.write(task_data)

    result = task_definitions.generate_task_definition(
        filename.strpath,
        environment={},
        template_vars={
            'image': 'my-docker-image:1.0',
        },
        overrides={},
        name='my-task-def',
    )
    expected = {
        'family': 'my-task-def',
        'volumes': [],
        'containerDefinitions': [
            {
                'name': 'default',
                'image': 'my-docker-image:1.0',
                'essential': True,
                'command': ['hello', 'world'],
                'hostname': 'my-task-def',
                'memory': 256,
                'cpu': 0,
                'portMappings': [
                    {
                        'containerPort': 8080,
                        'hostPort': 0
                    }
                ],
                'environment': {}
            }
        ]
    }
    assert result == expected


def test_generate_task_definition_overrides(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "default",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()

    filename = tmpdir.join('task_definition.json')
    filename.write(task_data)

    result = task_definitions.generate_task_definition(
        filename.strpath,
        environment={},
        template_vars={
            'image': 'my-docker-image:1.0',
        },
        overrides={
            'default': {
                'memory': 512,
                'memoryReservation': 128,
                'portMappings': [
                    {
                        'hostPort': 80,
                        'containerPort': 9000,
                    }
                ],
            }
        },
        name='my-task-def',
    )
    expected = {
        'family': 'my-task-def',
        'volumes': [],
        'containerDefinitions': [
            {
                'name': 'default',
                'image': 'my-docker-image:1.0',
                'essential': True,
                'command': ['hello', 'world'],
                'hostname': 'my-task-def',
                'memory': 512,
                'memoryReservation': 128,
                'cpu': 0,
                'portMappings': [
                    {
                        'containerPort': 8080,
                        'hostPort': 0
                    },
                    {
                        'containerPort': 9000,
                        'hostPort': 80
                    }
                ],
                'environment': {}
            }
        ]
    }
    assert result == expected


def test_generate_multiple_task_definitions(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "web-1",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        },
        {
          "name": "web-2",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()

    filename = tmpdir.join('task_definition.json')
    filename.write(task_data)

    config = {
        'environment': {
            'DATABASE_URL': 'postgresql://',
        },
        'environment_groups': {
            'group-1': {
                'ENV_CODE': 'group-1',
            },
            'group-2': {
                'ENV_CODE': 'group-2',
            },
        },
        'task_definitions': {
            'task-def-1': {
                'template': filename.strpath,
                'environment_group': 'group-1',
                'overrides': {
                    'web-1': {
                        'memory': 512,
                    }
                }
            },
            'task-def-2': {
                'template': filename.strpath,
                'environment_group': 'group-2',
                'overrides': {
                    'web-1': {
                        'memory': 512,
                    }
                }
            }
        }
    }

    result = task_definitions.generate_task_definitions(
        config,
        template_vars={
            'image': 'my-docker-image:1.0',
        },
        base_path=None)

    expected = {
        'task-def-1': {
            'definition': {
                'family': 'task-def-1',
                'volumes': [],
                'containerDefinitions': [
                    {
                        'name': 'web-1',
                        'image': 'my-docker-image:1.0',
                        'essential': True,
                        'command': ['hello', 'world'],
                        'hostname': 'task-def-1-web-1',
                        'memory': 512,
                        'cpu': 0,
                        'portMappings': [
                            {
                                'containerPort': 8080,
                                'hostPort': 0
                            }
                        ],
                        'environment': {
                            'DATABASE_URL': 'postgresql://',
                            'ENV_CODE': 'group-1',
                        }
                    },
                    {
                        'name': 'web-2',
                        'image': 'my-docker-image:1.0',
                        'essential': True,
                        'command': ['hello', 'world'],
                        'hostname': 'task-def-1-web-2',
                        'memory': 256,
                        'cpu': 0,
                        'portMappings': [
                            {
                                'containerPort': 8080,
                                'hostPort': 0
                            }
                        ],
                        'environment': {
                            'DATABASE_URL': 'postgresql://',
                            'ENV_CODE': 'group-1',
                        }
                    }
                ]
            },
        },
        'task-def-2': {
            'definition': {
                'family': 'task-def-2',
                'volumes': [],
                'containerDefinitions': [
                    {
                        'name': 'web-1',
                        'image': 'my-docker-image:1.0',
                        'essential': True,
                        'hostname': 'task-def-2-web-1',
                        'command': ['hello', 'world'],
                        'memory': 512,
                        'cpu': 0,
                        'portMappings': [
                            {
                                'containerPort': 8080,
                                'hostPort': 0
                            }
                        ],
                        'environment': {
                            'DATABASE_URL': 'postgresql://',
                            'ENV_CODE': 'group-2',
                        }
                    },
                    {
                        'name': 'web-2',
                        'image': 'my-docker-image:1.0',
                        'hostname': 'task-def-2-web-2',
                        'essential': True,
                        'command': ['hello', 'world'],
                        'memory': 256,
                        'cpu': 0,
                        'portMappings': [
                            {
                                'containerPort': 8080,
                                'hostPort': 0
                            }
                        ],
                        'environment': {
                            'DATABASE_URL': 'postgresql://',
                            'ENV_CODE': 'group-2',
                        }
                    }
                ]
            }
        }
    }

    assert result == expected


def test_generate_task_definitions_write_output(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "web-1",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "logConfiguration": {},
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()
    base_path = tmpdir.join('input').mkdir()

    filename = base_path.join('task_definition.json')
    filename.write(task_data)

    config = {
        'environment': {
            'DATABASE_URL': 'postgresql://',
        },
        'task_definitions': {
            'task-def-1': {
                'template': 'task_definition.json',
                'overrides': {
                    'web-1': {
                        'memory': 512,
                        "logConfiguration": {
                            "logDriver": "awslogs",
                            "options": {
                                "awslogs-group": "default",
                                "awslogs-region": "eu-west-1"
                            }
                        }
                    }
                }
            }
        }
    }

    task_definitions.generate_task_definitions(
        config,
        template_vars={
            'image': 'my-docker-image:1.0',
        },
        base_path=base_path.strpath,
        output_path=tmpdir.join('output').mkdir().strpath)

    assert tmpdir.join('output').join('task-def-1.json').exists()


def test_generate_task_definition_with_task_role_arn(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "default",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()

    filename = tmpdir.join('task_definition.json')
    filename.write(task_data)

    result = task_definitions.generate_task_definition(
        filename.strpath,
        environment={},
        task_role_arn="arn:my-task-role",
        template_vars={
            'image': 'my-docker-image:1.0',
        },
        overrides={},
        name='my-task-def',
    )

    expected = {
        'family': 'my-task-def',
        'taskRoleArn': 'arn:my-task-role',
        'volumes': [],
        'containerDefinitions': [
            {
                'name': 'default',
                'image': 'my-docker-image:1.0',
                'essential': True,
                'hostname': 'my-task-def',
                'command': ['hello', 'world'],
                'memory': 256,
                'cpu': 0,
                'portMappings': [
                    {
                        'containerPort': 8080,
                        'hostPort': 0
                    }
                ],
                'environment': {}
            }
        ]
    }
    assert result == expected



import logging

# Initialize logging
logger = logging.getLogger('deploy')
logger.setLevel(logging.DEBUG)

handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
handler.setFormatter(
    logging.Formatter('[%(asctime)s] %(message)s', datefmt='%H:%M:%S'))
logger.addHandler(handler)



#!/usr/bin/env python
import datetime
import copy
import operator
import os.path
import re
import sys
import time
import tokenize

import boto3
import click
import pytz
import yaml

from ecs_deplojo import utils
from ecs_deplojo.exceptions import DeploymentFailed
from ecs_deplojo.logger import logger
from ecs_deplojo.task_definitions import generate_task_definitions

POLL_TIME = 2


class VarType(click.ParamType):
    name = 'var'
    re_pattern = re.compile('^%s$' % tokenize.Name)

    def convert(self, value, param, ctx):
        try:
            key, value = value.split('=', 1)
            if not self.re_pattern.match(key):
                self.fail('%s is not a valid identifier' % key)
        except ValueError:
            self.fail('%s is not a valid key/value string' % value, param, ctx)

        return (key, value)


class Connection(object):
    def __init__(self, role_arn):
        credentials = {}
        if role_arn:
            sts = boto3.client('sts')
            resp = sts.assume_role(
                RoleArn=role_arn,
                RoleSessionName='ecs-deplojo')
            credentials.update({
                'aws_secret_access_key': resp['Credentials']['SecretAccessKey'],
                'aws_access_key_id': resp['Credentials']['AccessKeyId'],
                'aws_session_token': resp['Credentials']['SessionToken']
            })
        self.ecs = boto3.client('ecs', **credentials)


@click.command()
@click.option('--config', required=True, type=click.File())
@click.option('--var', multiple=True, type=VarType())
@click.option('--dry-run', is_flag=True, default=False)
@click.option('--output-path', required=False, type=click.Path())
@click.option('--role-arn', required=False, type=str)
def cli(config, var, output_path, dry_run, role_arn=None):
    base_path = os.path.dirname(config.name)
    config = yaml.load(config)
    template_vars = dict(var)

    connection = Connection(role_arn)
    cluster_name = config['cluster_name']
    services = config['services']
    logger.info(
        "Starting deploy on cluster %s (%s services)",
        cluster_name, len(services))


    # Generate the task definitions
    task_definitions = generate_task_definitions(
        config, template_vars, base_path, output_path)

    # Check if all task definitions required by the services exists
    for service_name, service in services.items():
        if service['task_definition'] not in task_definitions:
            logger.error(
                "Missing task definition %r for service %r",
                service['task_definition'], service_name)

    # Run the deployment
    if not dry_run:
        try:
            start_deployment(config, connection, task_definitions)
        except DeploymentFailed:
            sys.exit(1)

    sys.exit(0)


def start_deployment(config, connection, task_definitions):
    """Start the deployment.

    The following steps are executed:

    1. The task definitions are registered with AWS
    2. The before_deploy tasks are started
    3. The services are updated to reference the last task definitions
    4. The client poll's AWS until all deployments are finished.
    5. The after_deploy tasks are started.

    """
    cluster_name = config['cluster_name']
    services = config['services']

    # Register the task definitions in ECS
    register_task_definitions(connection, task_definitions)

    # Run tasks before deploying services
    tasks_before_deploy = config.get('before_deploy', [])
    run_tasks(
        connection, cluster_name, task_definitions, tasks_before_deploy)

    # Check if all services exist
    existing_services = utils.describe_services(
        connection.ecs, cluster=cluster_name,
        services=task_definitions.keys())
    available_services = {
        service['serviceName'] for service in existing_services
    }
    new_services = set(task_definitions.keys()) - available_services

    # Update services
    for service_name, service in services.items():
        task_definition = task_definitions[service['task_definition']]
        if service_name in new_services:
            logger.info(
                "Creating new service %s with task definition %s",
                service_name, task_definition['name'])
            connection.ecs.create_service(
                cluster=cluster_name,
                serviceName=service_name,
                desiredCount=1,
                taskDefinition=task_definition['name'])
        else:
            logger.info(
                "Updating service %s with task definition %s",
                service_name, task_definition['name'])
            connection.ecs.update_service(
                cluster=cluster_name,
                service=service_name,
                taskDefinition=task_definition['name'])

    is_finished = wait_for_deployments(
        connection, cluster_name, services.keys())

    if not is_finished:
        raise DeploymentFailed("Timeout")

    # Run tasks after deploying services
    tasks_after_deploy = config.get('after_deploy', [])
    run_tasks(
        connection, cluster_name, task_definitions, tasks_after_deploy)

    # Deregister task definitions
    deregister_task_definitions(connection, task_definitions)


def wait_for_deployments(connection, cluster_name, service_names):
    """Poll ECS until all deployments are finished (status = PRIMARY)

    """
    logger.info("Waiting for deployments")
    start_time = time.time()

    def service_description(service):
        """Return string in format of 'name (0/2)'"""
        name = service['serviceName']
        for deployment in service['deployments']:
            if deployment.get('status') != 'PRIMARY':
                continue

            desired = deployment['desiredCount']
            pending = deployment['pendingCount']
            running = deployment['runningCount']

            return '%s (%s/%s)' % (name, pending + running, desired)
        return name

    # Wait till all service updates are deployed
    time.sleep(5)

    utc_timestamp = (
        datetime.datetime.utcnow().replace(tzinfo=pytz.utc) -
        datetime.timedelta(seconds=5))
    last_event_timestamps = {name: utc_timestamp for name in service_names}
    logged_message_ids = set()
    ready_timestamp = None
    last_message = datetime.datetime.now()

    while True:
        services = utils.describe_services(
            connection.ecs, cluster=cluster_name, services=service_names)

        in_progress = [s for s in services if len(s['deployments']) > 1]

        messages = extract_new_event_messages(
            services, last_event_timestamps, logged_message_ids)
        for message in messages:
            logger.info(
                "%s - %s",
                message['createdAt'].strftime('%H:%M:%S'), message['message'])
            last_message = datetime.datetime.now()

        # So we haven't printed something for a while, let's give some feedback
        if last_message < datetime.datetime.now() - datetime.timedelta(seconds=10):
            logger.info(
                "Still waiting for: %s",
                ', '.join([s['serviceName'] for s in in_progress]))

        # 5 Seconds after the deployment is no longer in progress we mark it
        # as done.
        offset = datetime.datetime.utcnow() - datetime.timedelta(seconds=5)
        if ready_timestamp and offset > ready_timestamp:
            logger.info(
                "Deployment finished: %s",
                ', '.join([service_description(s) for s in services]))
            break

        # Set is_ready after the previous check so that we can wait for x
        # more seconds before ending the operation successfully.
        if not in_progress:
            ready_timestamp = datetime.datetime.utcnow()

        time.sleep(5)
        if time.time() - start_time > (60 * 15):
            logger.error("Giving up after 15 minutes")
            return False
    return True


def extract_new_event_messages(services, last_timestamps, logged_message_ids):
    for service in services:
        events = []
        for event in service['events']:
            if event['createdAt'] > last_timestamps[service['serviceName']]:
                events.append(event)

        for event in reversed(events):
            if event['id'] not in logged_message_ids:
                yield event
                logged_message_ids.add(event['id'])

        # Keep track of the timestamp of the last event
        if events:
            last_timestamps[service['serviceName']] = events[-1]['createdAt']


def run_tasks(connection, cluster_name, task_definitions, tasks):
    """Run one-off tasks.


    :parameter connection: The internal connection object.
    :type connection: Connection
    :parameter cluster_name: The cluster name to run the task on
    :type cluster_name: str
    :parameter task_definitions: dict of task definitions.
    :type task_definitions: dict
    :parameter tasks: list of tasks to run.
    :type tasks: list

    """
    num = 0

    for task in tasks:
        task_def = task_definitions[task['task_definition']]
        logger.info(
            "Starting one-off task '%s' via %s (%s)",
            task['command'], task_def['name'], task['container'])

        response = connection.ecs.run_task(
            cluster=cluster_name,
            taskDefinition=task_def['name'],
            overrides={
                'containerOverrides': [
                    {
                        'name': task['container'],
                        'command': task['command'].split(),
                    }
                ]
            },
            startedBy='ecs-deplojo',
            count=1
        )
        if response.get('failures'):
            logger.error(
                "Error starting one-off task: %r", response['failures'])

            # If we already started one task then we keep retrying until
            # the previous task is finished.
            if num > 0 and num <= 30:
                time.sleep(5)
            else:
                sys.exit(1)
        num += 1


def register_task_definitions(connection, task_definitions):
    """Update task definitions"""

    for service_name, values in task_definitions.items():
        definition = transform_definition(values['definition'])
        result = connection.ecs.register_task_definition(**definition)

        values['family'] = result['taskDefinition']['family']
        values['revision'] = result['taskDefinition']['revision']
        values['name'] = '%s:%s' % (
            result['taskDefinition']['family'],
            result['taskDefinition']['revision'])
        values['arn'] = result['taskDefinition']['taskDefinitionArn']
        logger.info(
            "Registered new task definition %s", values['name'])


def deregister_task_definitions(connection, task_definitions):
    """Deregister all task definitions not used currently"""

    def yield_arns(family):
        for page in paginator.paginate(familyPrefix=family):
            for arn in page['taskDefinitionArns']:
                yield arn

    logger.info("Deregistering old task definitions")
    for service_name, values in task_definitions.items():
        logger.info(" - %s", values['family'])

        paginator = connection.ecs.get_paginator('list_task_definitions')
        num = 0

        for arn in yield_arns(values['family']):
            num += 1
            if arn != values['arn']:
                connection.ecs.deregister_task_definition(taskDefinition=arn)

            if num > 10:
                break


def transform_definition(definition):
    result = copy.deepcopy(definition)
    for container in result['containerDefinitions']:
        container['environment'] = sorted([
            {'name': k, 'value': str(v)}
            for k, v in container['environment'].items()
        ], key=operator.itemgetter('name'))

    return result



def slice_iterable(iterable, size):
    """Slice an iterable in chunks of `size` length"""
    buffer = []
    for item in iterable:
        buffer.append(item)
        if len(buffer) == size:
            # yield collected full chunk
            yield buffer
            buffer = []
    if buffer:
        yield buffer


def describe_services(ecs, cluster, services):
    """Custom version since the default limits to 10 services per call"""

    result = []
    for services_slice in slice_iterable(services, 10):

        response = ecs.describe_services(
            cluster=cluster, services=services_slice)
        result.extend(response['services'])
    return result



import json
import os.path
from string import Template


def generate_task_definitions(config, template_vars, base_path,
                              output_path=None):
    """Generate the task definitions

    :parameter config: The yaml config contents
    :parameter template_vars: Key-Value dict with template replacements
    :parameter base_path: The base path (location of the config file)
    :parameter output_path: Optional path to write the task definitions to.
    :rtype dict:

    """
    task_definitions = {}

    for name, info in config['task_definitions'].items():
        # Default environment. Always create a new dict instance since it is
        # mutated.
        env_items = {}
        env_items.update(config.get('environment', {}))

        # Environment groups
        env_group = info.get('environment_group')
        if env_group:
            env_items.update(config['environment_groups'][env_group])

        overrides = info.get('overrides', {})
        definition = generate_task_definition(
            info['template'],
            env_items,
            template_vars,
            overrides,
            name=name,
            base_path=base_path,
            task_role_arn=info.get('task_role_arn'))

        if output_path:
            write_task_definition(name, definition, output_path)
        task_definitions[name] = {
            'definition': definition,
        }
    return task_definitions


def generate_task_definition(filename, environment, template_vars, overrides,
                             name, base_path=None, task_role_arn=None):

    """Generate the task definitions"""
    if base_path:
        filename = os.path.join(base_path, filename)

    with open(filename, 'r') as fh:
        data = json.load(fh)

        data['family'] = name
        if task_role_arn:
            data['taskRoleArn'] = task_role_arn

        num_containers = len(data['containerDefinitions'])
        for container in data['containerDefinitions']:
            container['image'] = Template(container['image']).substitute(template_vars)
            container['environment'] = environment

            hostname = name
            if num_containers > 1:
                hostname += '-%s' % container['name'].replace('_', '-')
            container.setdefault('hostname', hostname)

            if overrides:
                container_overrides = overrides.get(container['name'], {})
                for key, value in container_overrides.items():
                    if key in container and isinstance(container[key], list):
                        container[key].extend(value)
                    elif key in container and isinstance(container[key], dict):
                        container[key].update(value)
                    else:
                        container[key] = value

        return data


def write_task_definition(name, definition, output_path):
    filename = os.path.join(output_path, '%s.json' % name)
    with open(filename, 'w') as fh:
        json.dump(definition, fh, indent=4)
